@@ -1,13 +1,14 @@
 package Image::CairoSVG;
 use warnings;
 use strict;
+use utf8;
 
-our $VERSION = '0.13';
+our $VERSION = '0.14';
 
 # Core modules
 use Carp qw/carp croak/;
 use ExtUtils::ParseXS::Utilities 'trim_whitespace';
-use Math::Trig;
+use Math::Trig qw!acos pi rad2deg deg2rad!;
 use Scalar::Util 'looks_like_number';
 
 # Installed modules
@@ -80,11 +81,6 @@ sub render
     my ($self, $file) = @_;
     my $p = XML::Parser->new (
 	Handlers => {
-
-	    # I think (may be wrong) we only need to handle "start"
-	    # tags for SVG. As far as I know, everything in SVG is a
-	    # "start" tag plus attributes.
-
 	    Start => sub {
 		handle_start ($self, @_);
 	    },
@@ -115,11 +111,11 @@ sub render
 
 sub handle_end
 {
-    my ($self, $parser, $tag) = @_;
-    # At the end of a group, delete its attributes.
-    if ($tag eq 'g') {
-	delete $self->{attr};
-    }
+    # Last argument is $tag
+    my ($self, $parser, undef) = @_;
+    my $element = pop @{$self->{elements}};
+    my $attr = $element->{attr};
+    $self->do_fill_stroke ($attr);
 }
 
 # <svg> tag seen
@@ -137,6 +133,13 @@ sub svg
 	$height = $attr{height};
 	$height = svg_units ($height);
     }
+    if ($attr{fill}) {
+	$self->{fill} = $attr{fill};
+    }
+    if ($attr{stroke}) {
+	$self->{stroke} = $attr{stroke};
+    }
+
 
     # Use viewBox attribute
 
@@ -172,6 +175,7 @@ sub svg
 	$self->{surface} = $surface;
 	$self->make_cr ();
     }
+    $self->do_svg_attr (%attr);
 }
 
 # Start tag handler for the XML parser. This is private.
@@ -180,6 +184,30 @@ sub handle_start
 {
     my ($self, $parser, $tag, %attr) = @_;
 
+    my $parent = $self->{elements}[-1];
+    if ($parent) {
+	my $pattr = $parent->{attr};
+	for my $key (qw!
+	    fill
+	    stroke
+	    stroke-linecap
+	    stroke-linejoin
+	    stroke-width
+	!) {
+	    # So where were the spiders
+	    # While the fly tried to break our balls
+	    if ($pattr->{$key} && ! $attr{$key}) {
+		$attr{$key} = $pattr->{$key};
+	    }
+	}
+    }
+
+    my $element = {
+	tag => $tag,
+	attr => \%attr,
+    };
+    push @{$self->{elements}}, $element;
+
     if ($tag eq 'svg') {
 	$self->svg (%attr);
     }
@@ -212,18 +240,22 @@ sub handle_start
     }
     else {
 	if ($self->{verbose}) {
-	    print "Unhandled tag '$tag'.\n";
+	    # There are probably many of these since this module is
+	    # not up to spec, so only complain if the user wants
+	    # "verbose" messages.
+	    carp "Unhandled SVG tag '$tag'";
 	}
-	#	warn "Unknown tag '$tag' in $self->{file}";
     }
 
     # http://www.princexml.com/doc/7.1/svg/
     # g, rect, circle, ellipse, line, path, text, tspan
+    # Also "use" etc.
 }
 
 sub g
 {
     my ($self, %attr) = @_;
+    # Group element
 }
 
 sub rect
@@ -340,7 +372,8 @@ sub path
 {
     my ($self, %attr) = @_;
 
-    # Get and parse the "d" attribute from the path.
+    # Get and parse the "d" attribute from the path using
+    # Image::SVG::Path.
 
     my $d = $attr{d};
     croak "No d in path" unless $d;
@@ -382,9 +415,11 @@ sub path
 	    $cr->line_to (@{$element->{point}});
 	}
 	elsif ($key eq 'C') {
-	    $cr->curve_to (@{$element->{control1}},
-			   @{$element->{control2}},
-			   @{$element->{end}});
+	    $cr->curve_to (
+		@{$element->{control1}},
+		@{$element->{control2}},
+		@{$element->{end}},
+	    );
 	}
 	elsif ($key eq 'Z') {
 	    $cr->close_path ();
@@ -415,75 +450,172 @@ sub path
 }
 
 # This is a Perl translation of 
-# https://github.com/Kozea/CairoSVG/blob/74701790b5fd299e99f993b18ea676f3284907b4/cairosvg/surface/path.py
+# https://www.w3.org/TR/SVG/implnote.html#ArcImplementationNotes
 
 sub svg_arc
 {
     my ($self, $element) = @_;
+    my $cr = $self->{cr};
+    # Radii
     my $rx = $element->{rx};
     my $ry = $element->{ry};
-    my $x3 = $element->{x};
-    my $y3 = $element->{y};
-
-    my $cr = $self->{cr};
+    # End points
+    my $x2 = $element->{x};
+    my $y2 = $element->{y};
 
     # rx=0 or ry=0 means straight line
-
     if ($rx == 0 || $ry == 0) {
-	$cr->line_to ($x3, $y3);
+	$self->msg ("Arc has a zero radius rx=$rx or ry=$ry, treating as straight line");
+	$cr->line_to ($x2, $y2);
 	return;
     }
-
+    my $fa = $element->{large_arc_flag};
+    my $fs = $element->{sweep_flag};
+    if ($fa != 0 && $fa != 1) {
+	croak "large-arc-flag must be either 0 or 1";
+    }
+    if ($fs != 0 && $fs != 1) {
+	croak "sweep-flag must be either 0 or 1";
+    }
+    $self->msg ("A: inputs: large-arc-flag: $fa, sweep-flag: $fs");
+    # Start points
     my ($x1, $y1) = $cr->get_current_point ();
+    $self->msg ("A: inputs: arc start: ($x1, $y1)");
+    $self->msg ("A: inputs: arc end: ($x2, $y2)");
+    $self->msg ("A: inputs: radii: ($rx, $ry)");
+    my $phi = deg2rad ($element->{x_axis_rotation});
+    $self->msg ("A: inputs: φ = $phi radians");
+    my ($xd, $yd) = (($x1-$x2)/2, ($y1-$y2)/2);
+    #    $self->msg ("Midpoint of vector from end to start: ($xd, $yd)");
+    my $s = sin $phi;
+    my $c = cos $phi;
+    #    $self->msg ("sin φ = $s, cos φ = $c");
+    # Eq. 5.1
+    my ($x1d, $y1d) = ($xd * $c + $yd * $s, - $xd * $s + $yd * $c);
+    $self->msg ("Rotated midpoint: x1' = $x1d, y1' = $y1d");
+    my $factor;
+    my $lambda = ($x1d/$rx)**2 + ($y1d/$ry)**2;
+    if ($lambda > 1) {
+	$self->msg ("$lambda > 1, increasing radii");
+	my $sqrtlambda = sqrt ($lambda);
 
-    # Calculations
-
-    my $x_axis_rotation = (pi * $element->{x_axis_rotation})/180;
-    my $large_arc_flag = $element->{large_arc_flag};
-    my $sweep_flag = $element->{sweep_flag};
-
-    # Translate $x3, $y3 to relative coords
-
-    $x3 -= $x1;
-    $y3 -= $y1;
-
-    my $radii_ratio = $ry / $rx;
-
-    my ($xe, $ye) = rotate ($x3, $y3, -$x_axis_rotation);
-
-    $ye /= $radii_ratio;
-    my $angle = point_angle (0, 0, $xe, $ye);
-
-    $xe = sqrt ($xe**2 + $ye**2);
-    $ye = 0;
-
-    if ($xe / 2 > $rx) {
-	$rx = $xe / 2;
-    }
-    my $xc = $xe / 2;
-    my $yc = sqrt ($rx**2 - $xc**2);
-
-    if (! ($large_arc_flag ^ $sweep_flag)) {
-	$yc = -$yc;
-    }
-    ($xe, $ye) = rotate ($xe, 0, $angle);
-    ($xc, $yc) = rotate ($xc, $yc, $angle);
-
-    my $angle1 = point_angle ($xc, $yc, 0, 0);
-    my $angle2 = point_angle ($xc, $yc, $xe, $ye);
-
-    # Draw the arc
-
-    $cr->save ();
-    $cr->translate ($x1, $y1);
-    $cr->rotate ($x_axis_rotation);
-    if ($sweep_flag) {
-	$cr->arc ($xc, $yc, $rx, $angle1, $angle2);
+	$rx *= $sqrtlambda;
+	$ry *= $sqrtlambda;
+	$factor = 0;
     }
     else {
-	$cr->arc_negative ($xc, $yc, $rx, $angle1, $angle2);
+	my $den = ($rx * $y1d)**2 + ($ry * $x1d)**2;
+	my $num = ($rx * $ry)**2 - $den;
+	#    $self->msg ("den = $den, num = $num");
+	$factor = sqrt ($num / $den);
     }
-    $cr->restore ();
+    #    $self->msg ("factor = $factor");
+    my $sign = 1;
+    if ($fa == $fs) {
+	$sign = -1;
+    }
+    $factor *= $sign;
+    my $cxd =   $factor * $rx * $y1d / $ry;
+    my $cyd = - $factor * $ry * $x1d / $rx;
+    #    $self->msg ("A: transformed centre: ($cxd, $cyd)");
+    # Eq 5.3
+    my $cx = ($c * $cxd - $s * $cyd) + ($x1 + $x2) / 2;
+    my $cy = ($s * $cxd + $c * $cyd) + ($y1 + $y2) / 2;
+    $self->msg (sprintf ("A: centre of ellipse: (%.2f, %.2f)", $cx, $cy));
+    my @vec1 = (1,0);
+    # Eq. 5.5
+    my $xv2 = ($x1d - $cxd)/$rx;
+    my $yv2 = ($y1d - $cyd)/$ry;
+    my @vec2 = ($xv2, $yv2);
+    my $theta1 = vangle (\@vec1, \@vec2);
+    my $theta1d = rad2deg ($theta1);
+    $self->msg (sprintf ("Start angle θ1 = %.2f (%.2f°)", $theta1, $theta1d));
+    # Eq. 5.6
+    my $xv3 = (-$x1d - $cxd)/$rx;
+    my $yv3 = (-$y1d - $cyd)/$ry;
+    my @vec3 = ($xv3, $yv3);
+    #    $self->msg ("vec2 = @vec2");
+    #    $self->msg ("vec3 = @vec3");
+    my $dt = vangle (\@vec2, \@vec3);
+    my $dtd = rad2deg ($dt);
+    $self->msg ("Swept angle initially: Δθ = $dt ($dtd)");
+    if ($fs == 0) {
+
+	# if fS = 0 and the right side of (eq. 5.6) is greater than 0,
+	# then subtract 360°, whereas if fS = 1 and the right side of
+	# (eq. 5.6) is less than 0, then add 360°. In all other cases
+	# leave it as is.
+
+	if ($dt > 0) {
+	    $dt -= 2*pi;
+	}
+    }
+    elsif ($fs == 1) {
+	if ($dt < 0) {
+	    $dt += 2*pi;
+	}
+    }
+    $dtd = rad2deg ($dt);
+    $self->msg (sprintf ("Swept angle Δθ = %.2f (%.2f°)", $dt, $dtd));
+
+    if ($fs) {
+	$cr->arc ($cx, $cy, $rx, $theta1, $theta1+$dt);
+    }
+    else {
+	$cr->arc_negative ($cx, $cy, $rx, $theta1, $theta1+$dt);
+    }
+}
+
+# Helper for svg_arc
+
+# Eq. 5.4 of
+# https://www.w3.org/TR/SVG/implnote.html#ArcImplementationNotes
+
+sub vangle
+{
+    my ($u, $v) = @_;
+    my $ulen = vlen ($u);
+    my $vlen = vlen ($v);
+    my $sign;
+    my $vdot = vdot ($u, $v);
+    my $cross = vcross ($u, $v);
+    if ($cross == 0) {
+	if ($vdot < 0) {
+	    $sign = -1;
+	}
+	else {
+	    $sign = 1;
+	}
+    }
+    else {
+	$sign = $cross / abs ($cross);
+    }
+    my $value = $vdot / ($ulen * $vlen);
+    return $sign * acos ($value);
+}
+
+# Helper for vangle
+
+sub vdot
+{
+    my ($u, $v) = @_;
+    return $u->[0] * $v->[0] + $u->[1] * $v->[1];
+}
+
+# Helper for vangle
+
+sub vcross
+{
+    my ($u, $v) = @_;
+    return $u->[0] * $v->[1] - $u->[1] * $v->[0];
+}
+
+# Helper for vangle
+
+sub vlen
+{
+    my ($v) = @_;
+    return sqrt ($v->[0]**2 + $v->[1]**2);
 }
 
 # Quadratic bezier curve shim for Cairo
@@ -586,14 +718,6 @@ sub do_svg_attr
 	    $attr{$key} = $value;
 	}
     }
-    my $fill = $attr{fill};
-    if ($fill) {
-	trim_whitespace ($fill);
-    }
-    my $stroke = $attr{stroke};
-    if ($stroke) {
-	trim_whitespace ($stroke);
-    }
     my $cr = $self->{cr};
     my $stroke_width = $attr{"stroke-width"};
     if ($stroke_width) {
@@ -608,16 +732,44 @@ sub do_svg_attr
     if ($linejoin) {
 	$cr->set_line_join ($linejoin);
     }
+    my $fill = $attr{fill};
+    if ($fill) {
+	trim_whitespace ($fill);
+    }
+    if (! $fill) {
+	my $svgfill = $self->{svg}{fill};
+	if ($svgfill) {
+	    $fill = $svgfill;
+	}
+    }
+    my $stroke = $attr{stroke};
+    if ($stroke) {
+	trim_whitespace ($stroke);
+    }
+    my $fill_opacity = $attr{'fill-opacity'};
+    # Not sure how to handle this yet.
+    #    $self->do_fill_stroke ($cr, $fill, $stroke);
+}
+
+sub do_fill_stroke
+{
+    my ($self, $attr) = @_;
+    my $cr = $self->{cr};
+    my $fill = $attr->{fill};
+    my $stroke = $attr->{stroke};
 
     if ($fill && $fill ne 'none') {
 	if ($stroke && $stroke ne 'none') {
 	    $self->set_colour ($fill);
 	    $cr->fill_preserve ();
+	    $self->msg ("Filling with $fill");
 	    $self->set_colour ($stroke);
 	    $cr->stroke ();
+	    $self->msg ("Stroking with $stroke");
 	}
 	else {
 	    $self->set_colour ($fill);
+	    $self->msg ("Filling with $fill");
 	    $cr->fill ();
 	}
     }
@@ -626,6 +778,7 @@ sub do_svg_attr
 	$cr->stroke ();
     }
     elsif (! $fill && ! $stroke) {
+	$self->msg ("Filling with black");
 	# Fill with black seems to be the default.
 	$self->set_colour ('#000000');
 	$cr->fill ();
@@ -694,9 +847,7 @@ sub surface
     return $self->{surface};
 }
 
-# Return angle between x axis and point knowing given center.
-
-# https://github.com/Kozea/CairoSVG/blob/74701790b5fd299e99f993b18ea676f3284907b4/cairosvg/surface/helpers.py#L116
+# Direction of vector from ($cx, $cy) to ($px, $py) in radians
 
 sub point_angle
 {
@@ -704,11 +855,23 @@ sub point_angle
     return atan2 ($py - $cy, $px - $cx);
 }
 
+# Rotate $x and $y anticlockwise by $angle in radians
+
 sub rotate
 {
     my ($x, $y, $angle) = @_;
-    return ($x * cos ($angle) - $y * sin ($angle),
-	    $y * cos ($angle) + $x * sin ($angle));
+    my $s = sin $angle;
+    my $c = cos $angle;
+    return ($x * $c - $y * $s, $x * $s + $y * $c);
+}
+
+sub msg
+{
+    my ($self, $msg) = @_;
+    if (! $self->{verbose}) {
+	return;
+    }
+    print "$msg\n";
 }
 
 sub debugmsg
